/**
 * Copyright (c) 2009-2010 Mark Logic Corporation. All rights reserved.
 */
package com.marklogic.cpox;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.Authenticator;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.PasswordAuthentication;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.Iterator;
import java.util.Map;

/**
 * @author Michael Blakeley, Mark Logic Corporation
 * 
 */
public class HttpWorker extends Worker {

    protected String baseUrl;

    /**
     * @param controller
     * @param index
     */
    public HttpWorker(Controller controller, int index) {
        super(controller, index);

        baseUrl = config.getConnectionString();

        // extract login from base URL
        // this works with basic and digest authentication
        URL url;
        try {
            url = new URL(baseUrl);
            final String[] auth = url.getAuthority().split("@")[0]
                    .split(":");
            Authenticator.setDefault(new Authenticator() {
                protected PasswordAuthentication getPasswordAuthentication() {
                    return new PasswordAuthentication(auth[0], auth[1]
                            .toCharArray());
                }
            });
        } catch (MalformedURLException e) {
            throw new FatalException(baseUrl, e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.marklogic.cpox.Worker#bookmarkAction(java.lang.String)
     */
    @Override
    protected InputStream bookmarkAction(String user, String id)
            throws ActionException {
        String urlString;
        try {
            urlString = baseUrl + "/bookmark.xqy?uid=" + user + "&did="
                    + encode(id);
        } catch (Exception e) {
            throw new ActionException(user + ", " + id, e);
        }
        return doRequest(urlString);
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.marklogic.cpox.Worker#searchAction(java.lang.String[],
     * java.lang.String, int, java.util.Map)
     */
    @Override
    protected InputStream searchAction(String[] query, String language,
            int start, Map<String, String> facets) throws ActionException {
        StringBuilder urlString = new StringBuilder(baseUrl);
        urlString.append("/search.xqy?").append("s=").append(start)
                .append("&lang=").append(language).append("&q=").append(
                        Utilities.join(query, "+"));
        // add any facet parameters
        if (!facets.isEmpty()) {
            Iterator<String> keys = facets.keySet().iterator();
            String key = null;
            String value = null;
            try {
                String quote = "\"";
                while (keys.hasNext()) {
                    key = keys.next();
                    value = facets.get(key);
                    // some facet values may need quotes for whitespace
                    // NB - timestamp is bucketed, so can't be quoted.
                    quote = key.equals("timestamp") ? "" : "\"";
                    urlString.append("+").append(key).append(":").append(
                            quote).append(encode(value)).append(quote);
                }
            } catch (Exception e) {
                throw new ActionException(key + "=" + value, e);
            }
        }
        return doRequest(urlString.toString());
    }

    /**
     * @param urlString
     * @return
     * @throws ActionException
     */
    private InputStream doRequest(String urlString)
            throws ActionException {
        URL url;
        HttpURLConnection conn = null;
        try {
            url = new URL(urlString);
            conn = (HttpURLConnection) url.openConnection();
            // use keep-alive to reduce open sockets
            conn.setRequestProperty("Connection", "keep-alive");
            // TODO: SSL support
            conn.setRequestMethod("GET");
            conn.setDoOutput(true);
            return conn.getInputStream();
        } catch (FileNotFoundException e) {
            // HTTP error
            try {
                throw new ActionException(urlString
                        + ": "
                        + ((null == conn) ? "null" : conn
                                .getResponseMessage()), e, false);
            } catch (IOException e1) {
                // now we are really in trouble
                throw new FatalException(e1);
            }
        } catch (MalformedURLException e) {
            throw new FatalException(urlString, e);
        } catch (IOException e) {
            throw new ActionException(urlString, e);
        }
    }

    /**
     * @param value
     * @return
     * @throws URISyntaxException
     * @throws UnsupportedEncodingException
     */
    private String encode(String value) throws URISyntaxException,
            UnsupportedEncodingException {
        // this seems awkward...
        if (null == value) {
            return null;
        }
        return URLEncoder.encode(value, "UTF-8");
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.marklogic.cpox.Worker#updateAction(java.lang.String,
     * java.lang.String)
     */
    @Override
    protected InputStream updateAction(String id, String token)
            throws ActionException {
        String urlString;
        try {
            urlString = baseUrl + "/update.xqy?id=" + encode(id)
                    + "&token=" + encode(token);
        } catch (Exception e) {
            throw new ActionException(id + ", " + token, e);
        }
        return doRequest(urlString);
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.marklogic.cpox.Worker#verifyAction(java.lang.String)
     */
    @Override
    protected InputStream verifyAction(String token)
            throws ActionException {
        String urlString;
        try {
            urlString = baseUrl + "/verify.xqy?token=" + encode(token);
        } catch (Exception e) {
            throw new ActionException(token, e);
        }
        return doRequest(urlString);
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.marklogic.cpox.Worker#viewAction(java.lang.String)
     */
    @Override
    protected InputStream viewAction(String id) throws ActionException {
        String urlString;
        try {
            urlString = baseUrl + "/view.xqy?id=" + encode(id);
        } catch (Exception e) {
            throw new ActionException(id, e);
        }
        return doRequest(urlString);
    }

}
