/**
 * Copyright (c) 2009-2010 Mark Logic Corporation. All rights reserved.
 */
package com.marklogic.cpox;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;

/**
 * @author Michael Blakeley, Mark Logic Corporation
 * 
 *         This abstract class describes a worker thread. Worker threads are
 *         created and run by a Controller.
 * 
 */
/**
 * @author Michael Blakeley, Mark Logic Corporation
 * 
 */
public abstract class Worker extends Thread {

    protected ResultLogger resultLogger;

    protected Configuration config;

    protected SimpleLogger logger;

    protected int index;

    protected Controller controller;

    protected TermGenerator generator;

    protected Random random;

    protected static int[] termSizes;

    protected static final boolean DEBUG = false;

    protected byte[] readBuffer = new byte[32 * 1024];

    protected long requests = 0;

    protected static enum ACTION {
        SEARCH, PAGINATE, FACET, VIEW, BOOKMARK, UPDATE, VERIFY
    };

    protected ACTION lastAction = null;

    protected String[] lastQuery = null;

    protected String lastQueryLanguage = null;

    protected Map<String, String> activeFacets = new HashMap<String, String>();

    protected Map<String, String[]> lastFacets = new HashMap<String, String[]>();

    protected int lastIndex;

    protected List<String> lastIds = new ArrayList<String>(10);

    protected int lastEnd;

    protected long lastCount;

    protected String token;

    protected Collection<String> bookmarkedIds = new HashSet<String>();

    protected Collection<String> updatedIds = new HashSet<String>();

    protected static Map<ACTION, ACTION[]> actionsMap = new HashMap<ACTION, ACTION[]>(
            3);

    static {
        // set up probabilities for actions
        ACTION[] actions;

        // last action was search (also facet, paginate)
        actions = new ACTION[100];
        actionSet(actions, 0, 5, ACTION.SEARCH);
        actionSet(actions, 5, 15, ACTION.FACET);
        actionSet(actions, 20, 5, ACTION.PAGINATE);
        actionSet(actions, 25, 75, ACTION.VIEW);
        actionsMap.put(ACTION.SEARCH, actions);

        // last action was view
        actions = new ACTION[50];
        actionSet(actions, 0, 30, ACTION.VIEW);
        actionSet(actions, 30, 10, ACTION.SEARCH);
        actionSet(actions, 40, 8, ACTION.UPDATE);
        actionSet(actions, 48, 2, ACTION.BOOKMARK);
        actionsMap.put(ACTION.VIEW, actions);

        // last action was update
        actions = new ACTION[10];
        actionSet(actions, 0, 6, ACTION.SEARCH);
        actionSet(actions, 6, 4, ACTION.VERIFY);
        actionsMap.put(ACTION.UPDATE, actions);

        // set up probability array for number of terms per SEARCH operation
        termSizes = new int[100];
        int start = 0;
        // 10% 1-term
        for (int i = start; i < 10; i++) {
            termSizes[i] = 1;
            start = 1 + i;
        }
        // 35% 2-term
        for (int i = start; i < 45; i++) {
            termSizes[i] = 2;
            start = 1 + i;
        }
        // 25% 3-term
        for (int i = start; i < 70; i++) {
            termSizes[i] = 3;
            start = 1 + i;
        }
        // 15% 4-term
        for (int i = start; i < 85; i++) {
            termSizes[i] = 4;
            start = 1 + i;
        }
        // 10% 5-term
        for (int i = start; i < 95; i++) {
            termSizes[i] = 5;
            start = 1 + i;
        }
        // 5% long tail, with 16-20 terms
        for (int i = start; i < 100; i++) {
            termSizes[i] = 16 + i - start;
        }

        if (DEBUG) {
            System.err.print("termSizes:");
            for (int i = 0; i < termSizes.length; i++) {
                System.err.print(" " + termSizes[i]);
            }
            System.err.println();
        }
    }

    /**
     * @param _controller
     * @param _index
     */
    public Worker(Controller _controller, int _index) {
        controller = _controller;
        index = _index;

        config = controller.getConfig();
        logger = config.getLogger();
        generator = controller.getGenerator();

        setName(controller.getName() + "-Worker-" + index);
        logger.info(getName());

        logger.info("logger level=" + logger.getLevel());
        logger.info("config level=" + config.getLogLevel());
    }

    /**
     * @param _array
     * @param _start
     * @param _length
     * @param _action
     */
    private static void actionSet(ACTION[] _array, int _start,
            int _length, ACTION _action) {
        int stop = _start + _length;
        if (_array.length < _start + _length) {
            throw new ArrayIndexOutOfBoundsException("" + _array.length
                    + " < " + _start + " + " + _length + " (" + stop
                    + ") for " + _action);
        }
        for (int i = _start; i < stop; i++) {
            _array[i] = _action;
        }
    }

    @Override
    public void run() {
        // simple timed test with random queries
        TimedEvent te;
        requests = 0;
        token = Integer.toHexString(getName().hashCode());
        logger.info(getName() + ", token " + token);
        random = new Random(controller.getName().hashCode() + index);

        try {
            resultLogger = new ResultLogger(config.getStart(), getName());
            logger.info("results in " + resultLogger.getPath());
        } catch (IOException e) {
            controller.reportException(e);
            tryClose(resultLogger);
            throw new FatalException(e);
        }

        long endNanos = System.nanoTime() + config.getTestTimeNanos();
        while (endNanos > System.nanoTime()) {
            te = new TimedEvent();
            try {
                resultLogger.log(te, nextAction(te));
            } catch (Exception e) {
                controller.reportException(e);
                try {
                    resultLogger.log(te, e);
                } catch (IOException ioe) {
                    // now we are really in trouble
                    tryClose(resultLogger);
                    throw new FatalException(ioe);
                }
                if (config.isFatalErrors()) {
                    tryClose(resultLogger);
                    throw new FatalException(e);
                }
            } finally {
                // count errors as requests, even with fatal errors
                requests++;
            }
        }

        tryClose(resultLogger);
        controller.interrupt();
        logger.info(getName() + " exiting");
    }

    /**
     * @param _c
     */
    protected void tryClose(Closeable _c) {
        try {
            _c.close();
        } catch (IOException e) {
            // log and ignore
            logger.logException(e);
        }
    }

    /**
     * @return
     */
    public long getRequestCount() {
        return requests;
    }

    /**
     * @return
     * @throws IOException
     */
    public Reader getResultsReader() throws IOException {
        return resultLogger.getReader();
    }

    /**
     * @param _te
     * @return
     * @throws ActionException
     */
    protected long nextAction(TimedEvent _te) throws ActionException {
        ACTION nextAction = (
        // bootstrap the first action
        (null == lastAction) ? ACTION.SEARCH
                : ((ACTION.FACET.equals(lastAction) || ACTION.PAGINATE
                        .equals(lastAction)) ? ACTION.SEARCH
                        : ((ACTION.BOOKMARK.equals(lastAction) || ACTION.VERIFY
                                .equals(lastAction)) ? ACTION.VIEW
                                : lastAction)));
        if (null != lastAction) {
            // look up probabilities for selected state
            ACTION[] actions = actionsMap.get(nextAction);
            if (null == actions) {
                throw new NullPointerException("null actions for "
                        + lastAction);
            }
            nextAction = actions[random.nextInt(actions.length)];
        }

        InputStream result = doAction(_te, nextAction);

        try {
            return process(_te, result);
        } catch (IOException e) {
            throw new ActionException(e);
        }
    }

    /**
     * @param _te
     * @param nextAction
     * @return
     * @throws ActionException
     */
    protected InputStream doAction(TimedEvent _te, ACTION nextAction)
            throws ActionException {
        // book keeping, so the caller knows how to process the results
        lastAction = (ACTION.FACET.equals(nextAction) || ACTION.PAGINATE
                .equals(nextAction)) ? ACTION.SEARCH : nextAction;

        InputStream result;
        switch (nextAction) {
        case BOOKMARK:
            result = nextActionBookmark(_te);
            break;
        case FACET:
            result = nextActionFacet(_te);
            break;
        case PAGINATE:
            result = nextActionPaginate(_te);
            break;
        case SEARCH:
            result = nextActionSearch(_te);
            break;
        case UPDATE:
            result = nextActionUpdate(_te);
            break;
        case VERIFY:
            result = nextActionVerify(_te);
            break;
        case VIEW:
            result = nextActionView(_te);
            break;
        default:
            throw new ActionException("unexpected action: " + nextAction);
        }
        return result;
    }

    /**
     * @param _action
     * @param _te
     * @param _result
     * @throws IOException
     * @throws ActionException
     */
    protected long process(ACTION _action, TimedEvent _te,
            InputStream _result) throws IOException, ActionException {
        return process(_te, _result);
    }

    /**
     * @param _te
     * @param _result
     * @throws IOException
     * @throws ActionException
     */
    protected long process(TimedEvent _te, InputStream _result)
            throws IOException, ActionException {
        // record time before response handling
        long nanos = System.nanoTime();

        // do not handle facet or paginate explicitly
        switch (lastAction) {
        case BOOKMARK:
            parseBookmark(_te, drainBytes(_te, _result, true));
            break;
        case SEARCH:
            parseResults(lastAction, _te, drainBytes(_te, _result, true));
            break;
        case VERIFY:
            verify(_te, drainBytes(_te, _result, true));
        default:
            drainBytes(_te, _result, false);
            break;
        }

        return nanos;
    }

    /**
     * @param te
     * @param result
     * @throws IOException
     * @throws NumberFormatException
     * @throws ActionException
     */
    protected void parseBookmark(TimedEvent te, InputStream result)
            throws NumberFormatException, IOException, ActionException {
        BufferedReader br = new BufferedReader(new InputStreamReader(
                result));
        int count = Integer.parseInt(br.readLine());
        if (count != bookmarkedIds.size()) {
            throw new ActionException(
                    "bookmark token count mismatch for " + token + ", "
                            + count + " != " + bookmarkedIds.size(),
                    false);
        }
        logger.fine("bookmark token count ok, " + token + " = " + count);
    }

    /**
     * @param te
     * @param result
     * @throws IOException
     * @throws NumberFormatException
     * @throws ActionException
     */
    protected void verify(TimedEvent te, InputStream result)
            throws NumberFormatException, IOException, ActionException {
        BufferedReader br = new BufferedReader(new InputStreamReader(
                result));
        int count = Integer.parseInt(br.readLine());
        if (count != updatedIds.size()) {
            throw new ActionException("update token count mismatch for "
                    + token + ", " + count + " != " + updatedIds.size(),
                    false);
        }
        logger.fine("update token count ok, " + token + " = " + count);
    }

    /**
     * @param action
     * @param te
     * @param result
     */
    protected void parseResults(ACTION action, TimedEvent te,
            InputStream result) throws ActionException {
        lastFacets.clear();
        lastIds.clear();

        // minimal validation
        // parse information from search results
        XmlPullParser xpp = null;
        try {
            // logger.fine("available = " + result.available());
            xpp = config.getXppFactory().newPullParser();
            xpp
                    .setFeature(XmlPullParser.FEATURE_PROCESS_NAMESPACES,
                            true);
            xpp.setInput(new InputStreamReader(result));

            int eventType;
            String name = null;
            String namespace = null;
            String facet = null;
            String id = null;
            List<String> values = new ArrayList<String>();
            int facetCount = 0;
            int lastFacetCount = 0;
            int resultCount = 0;

            do {
                eventType = xpp.next();
                switch (eventType) {
                case XmlPullParser.END_TAG:
                    name = xpp.getName();
                    // logger.fine("end tag " + name);
                    if ("facet".equals(name)) {
                        // logger.fine("facet " + facet + ", size = "
                        // + values.size());
                        if (values.size() > 0) {
                            lastFacets.put(facet, values
                                    .toArray(new String[0]));
                            values.clear();
                            // logger.fine(facet + " "
                            // + lastFacets.get(facet).length);
                        }
                        facet = null;
                        lastFacetCount = 0;
                    }
                    name = null;
                    break;
                case XmlPullParser.START_TAG:
                    name = xpp.getName();
                    if (null == namespace) {
                        namespace = xpp.getNamespace();
                    }
                    // logger.fine("start tag " + name);
                    if ("response".equals(name)) {
                        // populate lastIndex
                        // populate lastCount
                        lastCount = Long.parseLong(xpp.getAttributeValue(
                                null, "total"));
                        lastIndex = Integer.parseInt(xpp
                                .getAttributeValue(null, "start"));
                    } else if ("result".equals(name)) {
                        resultCount++;
                        // populate lastIds
                        id = xpp.getAttributeValue("", "uri");
                        if (null == id) {
                            throw new ActionException("null uri");
                        }
                        if (lastIds.contains(id)) {
                            throw new ActionException("duplicate id "
                                    + id);
                        }
                        logger.fine("id = " + id);
                        lastIds.add(id);
                    } else if ("facet".equals(name)) {
                        // populate key for facetMap
                        if (null != facet) {
                            throw new ActionException(
                                    "unexpected facet: " + facet);
                        }
                        facet = xpp.getAttributeValue(null, "name");
                    } else if ("facet-value".equals(name)) {
                        // populate values for facetMap
                        if (null == facet) {
                            throw new ActionException(
                                    "facet-value without facet");
                        }
                        facetCount = Integer.parseInt(xpp
                                .getAttributeValue(null, "count"));
                        if (facetCount < 1) {
                            throw new ActionException(
                                    "facet value failure for " + facet
                                            + ": 0");
                        }
                        // timestamp will be a histogram,
                        // therefore not in frequency order
                        if (!"timestamp".equals(facet)
                                && lastFacetCount > 0
                                && lastFacetCount < facetCount) {
                            throw new ActionException(
                                    String
                                            .format(
                                                    "facet ordering failure for %s: %d < %d; %s=%s",
                                                    facet,
                                                    lastFacetCount,
                                                    facetCount,
                                                    lastQueryLanguage,
                                                    lastQuery));
                        }
                        lastFacetCount = facetCount;
                        values.add(xpp.getAttributeValue(null, "name"));
                    }
                    break;
                default:
                    break;
                }
            } while (eventType != XmlPullParser.END_DOCUMENT);

            if (resultCount != lastIds.size()) {
                throw new ActionException("result mismatch: "
                        + resultCount + " != " + lastIds.size());
            }

            lastEnd = lastIndex + resultCount;

            // selecting a facet should always return results
            if (0 == lastCount && !activeFacets.isEmpty()) {
                throw new ActionException(
                        queryAsString("empty response with facets"));
            }

            // naturally an empty results page will have no facets
            if (lastCount > 0 && lastFacets.isEmpty()) {
                throw new ActionException(queryAsString("no facet data"));
            }
        } catch (XmlPullParserException e) {
            throw new FatalException(e);
        } catch (IOException e) {
            logger.warning("" + action);
            if (null != xpp) {
                logger.warning(e.getMessage() + " at "
                        + xpp.getPositionDescription());
            }
            throw new ActionException(e);
        } finally {
            tryClose(result);
        }
    }

    /**
     * @param msg
     * @return
     */
    private String queryAsString(String msg) {
        String message = String.format("%s for q=%s&lang=%s", msg,
                Utilities.join(lastQuery, "+"), lastQueryLanguage);
        Iterator<String> keys = activeFacets.keySet().iterator();
        String k;
        while (keys.hasNext()) {
            k = keys.next();
            message += String.format("&%s=%s", k, activeFacets.get(k));
        }
        return message;
    }

    /**
     * @param te
     * @param result
     * @param _copy
     * @return
     * @throws IOException
     */
    protected InputStream drainBytes(TimedEvent te, InputStream result,
            boolean _copy) throws IOException {
        // record bytes received
        InputStream is = null;
        ByteArrayOutputStream os = _copy ? new ByteArrayOutputStream()
                : null;
        try {
            int actual = 0;
            while ((actual = result.read(readBuffer)) > -1) {
                te.increment(actual);
                if (_copy) {
                    os.write(readBuffer, 0, actual);
                }
            }
        } finally {
            if (null != is) {
                is.close();
            }
        }
        return _copy ? new ByteArrayInputStream(os.toByteArray()) : null;
    }

    protected String nextId(boolean remove) {
        // return next id for bookmark, update, view
        // randomly select an id from the last page of search results
        // if no ids are available, return null
        if (null == lastIds || lastIds.isEmpty()) {
            // null signals caller to select a new action
            return null;
        }
        String id = lastIds.get(random.nextInt(lastIds.size()));
        if (remove) {
            lastIds.remove(id);
        }
        return id;
    }

    protected String nextId(Collection<String> observed) {
        String id = null;
        do {
            // set the remove flag - if we've updated everything,
            // we want to trigger another search.
            id = nextId(true);
            if (null == id) {
                break;
            }
            if (observed.contains(id)) {
                id = null;
            }
        } while (null == id);
        return id;
    }

    /**
     * @param te
     * @return
     * @throws ActionException
     */
    protected InputStream nextActionBookmark(TimedEvent te)
            throws ActionException {
        te.setDescription("bookmark");
        String id = nextId(bookmarkedIds);
        if (null == id) {
            logger.fine("unable to " + te.getDescription() + ": no data");
            return doAction(te, ACTION.SEARCH);
        }

        // make sure we don't update the token count unless successful
        InputStream result = bookmarkAction(token, id);
        bookmarkedIds.add(id);
        // logger.fine("bookmarks for token " + token + " = "
        // + bookmarkedIds.size());
        return result;
    }

    /**
     * @param te
     * @return
     * @throws ActionException
     */
    protected InputStream nextActionFacet(TimedEvent te)
            throws ActionException {
        te.setDescription("facet");
        // determine type and value from last search results
        Set<String> keySet = lastFacets.keySet();
        if (null == keySet) {
            // no facet data available from last search
            logger.warning("unable to facet: no key set data");
            return doAction(te, ACTION.SEARCH);
        }
        String[] types = keySet.toArray(new String[0]);
        if (types.length < 1) {
            // no facet data available from last search,
            // or we've used it all - time for another search!
            logger.fine("unable to facet: no types data");
            return doAction(te, ACTION.SEARCH);
        }
        String type = types[random.nextInt(types.length)];
        String[] values = lastFacets.get(type);
        if (types.length < 1) {
            // no facet data available from last search
            logger.warning("unable to facet: no values data for " + type);
            return doAction(te, ACTION.SEARCH);
        }
        String value = values[random.nextInt(values.length)];
        if (null == type || null == value) {
            // no facet data available from last search
            logger.warning("unable to facet: no data");
            return doAction(te, ACTION.SEARCH);
        }
        activeFacets.put(type, value);
        lastIndex = 1;
        return searchAction();
    }

    /**
     * @param te
     * @return
     * @throws ActionException
     */
    protected InputStream nextActionPaginate(TimedEvent te)
            throws ActionException {
        te.setDescription("paginate");
        // determine next start item from last search results
        int start = 1 + lastEnd;
        // check previous search results to see if we have more
        if (start > lastCount) {
            logger.fine("unable to paginate: " + start + " > "
                    + lastCount);
            return doAction(te, ACTION.SEARCH);
        }
        return searchAction(start);
    }

    /**
     * @param _te
     * @return
     * @throws ActionException
     */
    protected InputStream nextActionSearch(TimedEvent _te)
            throws ActionException {
        String[] query = new String[termSizes[random
                .nextInt(termSizes.length)]];
        _te.setDescription("search " + query.length);

        for (int i = 0; i < query.length; i++) {
            query[i] = generator.nextWord();
        }

        lastQuery = query;
        lastQueryLanguage = generator.getLanguage();
        lastIndex = 1;
        activeFacets.clear();
        return searchAction();
    }

    /**
     * @param te
     * @return
     * @throws ActionException
     */
    protected InputStream nextActionUpdate(TimedEvent te)
            throws ActionException {
        te.setDescription("update");
        String id = nextId(updatedIds);
        if (null == id) {
            logger.fine("unable to " + te.getDescription() + ": no data");
            return doAction(te, ACTION.SEARCH);
        }

        // make sure we don't update the token count unless successful
        InputStream result = updateAction(id, token);
        updatedIds.add(id);
        // logger.fine("updates for token " + token + " = "
        // + updatedIds.size() + ", " + id);
        return result;
    }

    /**
     * @param te
     * @return
     * @throws ActionException
     */
    protected InputStream nextActionVerify(TimedEvent te)
            throws ActionException {
        te.setDescription("verify");
        return verifyAction(token);
    }

    /**
     * @param te
     * @return
     * @throws ActionException
     */
    protected InputStream nextActionView(TimedEvent te)
            throws ActionException {
        te.setDescription("view");
        // it is ok if we view the same doc more than once,
        // as long as it happens within different search result sets
        String id = nextId(true);
        if (null == id) {
            logger.fine("unable to " + te.getDescription() + ": no data");
            return doAction(te, ACTION.SEARCH);
        }
        return viewAction(id);
    }

    /**
     * @param userId
     * @param documentId
     * @return
     * @throws ActionException
     */
    protected abstract InputStream bookmarkAction(String userId,
            String documentId) throws ActionException;

    /**
     * @param start
     * @return
     * @throws ActionException
     */
    protected InputStream searchAction(int start) throws ActionException {
        return searchAction(lastQuery, lastQueryLanguage, start,
                activeFacets);
    }

    /**
     * @return
     * @throws ActionException
     */
    protected InputStream searchAction() throws ActionException {
        return searchAction(lastQuery, lastQueryLanguage, lastIndex,
                activeFacets);
    }

    /**
     * @param query
     * @param start
     * @param facets
     * @return
     * @throws ActionException
     */
    protected abstract InputStream searchAction(String[] query,
            String queryLanguage, int start, Map<String, String> facets)
            throws ActionException;

    /**
     * @param id
     * @param token
     * @return
     * @throws ActionException
     */
    protected abstract InputStream updateAction(String id, String token)
            throws ActionException;

    /**
     * @return
     * @throws ActionException
     */
    protected abstract InputStream verifyAction(String token)
            throws ActionException;

    /**
     * @param id
     * @return
     * @throws ActionException
     */
    protected abstract InputStream viewAction(String id)
            throws ActionException;
}
